import json
import logging
from pathlib import Path
from typing import List, Dict, Callable
import io
import pandas as pd

from aito.utils.schema_handler import SchemaHandler


class DataFrameHandler:
    allowed_format = ['csv', 'json', 'excel', 'ndjson']

    def __init__(self):
        self.logger = logging.getLogger('DataFrameHandler')
        self.default_options = {
            'csv': {},
            'excel': {},
            'json': {'orient': 'records'},
            'ndjson': {'orient': 'records', 'lines': True}
        }
        self.default_apply_functions = [self.datetime_to_string]
        self.schema_handler = SchemaHandler()

    def validate_in_out_format(self, in_format: str, out_format: str):
        """
        Validate the file parameters of the converter
        :param in_format: input format
        :param out_format: output format
        """
        if in_format not in self.allowed_format:
            raise ValueError(f"Expect input format to be {str(self.allowed_format)} instead of {in_format}")

        if out_format not in self.allowed_format:
            raise ValueError(f"Expect output format to be {str(self.allowed_format)} instead of {out_format}")

    @staticmethod
    def datetime_to_string(df: pd.DataFrame) -> pd.DataFrame:
        # Aito hasn't support datetime data yet. Converting all datetime data to string
        for col in df:
            if df[col].dtypes == 'datetime64[ns]':
                df[col] = df[col].astype(str)
        return df

    @staticmethod
    def apply_functions_on_df(df: pd.DataFrame, functions: List[Callable]) -> pd.DataFrame:
        """
        Applying functions sequentially to a dataframe
        :param df:
        :param functions:
        :return:
        """
        for f in functions:
            df = f(df)
        return df

    def convert_df_from_aito_table_schema(self, df: pd.DataFrame, table_schema: Dict):
        self.schema_handler.validate_table_schema(table_schema)

        columns_schema = table_schema['columns']
        df_columns = set(df.columns.values)
        table_schema_columns = set(columns_schema.keys())

        for col in (df_columns - table_schema_columns):
            self.logger.warn(f"Column '{col}' found in the input data but not found in the input schema")
        for col in (table_schema_columns - df_columns):
            self.logger.warn(f"Column '{col}' found in the input schema but not found in the input data")

        for col in table_schema_columns.intersection(df_columns):
            col_schema = columns_schema[col]
            col_schema_nullable = True if ('nullable' not in col_schema or col_schema['nullable']) else False
            if not col_schema_nullable and df[col].isna().any():
                raise ValueError(f"Column '{col}' is nullable but stated non-nullable in the input schema")
            col_schema_dtypes = self.schema_handler.aito_types_to_pandas_dtypes[col_schema['type']]
            if df[col].dtype.name not in col_schema_dtypes:
                self.logger.info(f"Converting column '{col}' to {col_schema['type']} type according to the schema...")
                try:
                    df[col] = df[col].astype(col_schema_dtypes[0], skipna=True)
                except Exception as e:
                    self.logger.error(f'Conversion error: {e}')
                    raise e
        return df

    def read_file_to_df(self, read_input, in_format: str, read_options: Dict = None) -> pd.DataFrame:
        """
        Load a file and return pandas Dataframe
        :param read_input: path to or buffer of input
        :param in_format: input format
        :param read_options: dictionary contains arguments for pandas read function
        :return:
        """
        if isinstance(read_input, io.TextIOWrapper):
            self.logger.info("Start reading from standard input...")
        else:
            self.logger.info(f"Start reading input from {read_input}...")
        read_functions = {'csv': pd.read_csv, 'excel': pd.read_excel, 'json': pd.read_json, 'ndjson': pd.read_json}

        if not read_options:
            options = self.default_options[in_format]
        else:
            options = read_options
            options.update(self.default_options[in_format])
        df = read_functions[in_format](read_input, **options)
        return df

    def df_to_format(self, df: pd.DataFrame, out_format: str, write_output, convert_options: Dict = None):
        """

        :param df:
        :param out_format:
        :param write_output:
        :param convert_options: dictionary contains arguments for pandas convert function
        :return:
        """
        """

        :param df:
        :param out_format: file output format
        :param out_file_name: if output file name is not defined, it will be the same as input file name
        :param out_file_folder: path to output file folder
        :param convert_options: dictionary contains arguments for pandas convert function
        :return:
        """
        self.logger.info(f"Start converting to {out_format} and writing to output...")
        convert_functions = {'csv': df.to_csv, 'excel': df.to_excel, 'json': df.to_json, 'ndjson': df.to_json}
        if not convert_options:
            options = self.default_options[out_format]
        else:
            options = convert_options
            options.update(self.default_options[out_format])

        convert_functions[out_format](write_output, **options)

    def convert_file(self,
                     read_input,
                     write_output,
                     in_format: str,
                     out_format: str,
                     read_options: Dict = None,
                     convert_options: Dict = None,
                     apply_functions: List[Callable[..., pd.DataFrame]] = None,
                     create_table_schema: Path = None,
                     use_table_schema: Dict = None):
        """
        Converting a file into expected format and generate aito schema if required
        :param read_input: filepath to input or input buffer
        :param write_output: filepath to output or output buffer
        :param in_format: input format
        :param out_format: output format
        :param read_options: dictionary contains arguments for pandas read function
        :param convert_options: dictionary contains arguments for pandas convert function
        :param apply_functions: List of partial functions that will be chained applied to the loaded pd.DataFrame
        :param create_table_schema: option to auto generate aito schema
        :param use_table_schema: use an aito schema to dictates data types and convert the data
        :return:
        """
        self.validate_in_out_format(in_format, out_format)

        df = self.read_file_to_df(read_input, in_format, read_options)

        if apply_functions:
            apply_functions = self.default_apply_functions + apply_functions
        else:
            apply_functions = self.default_apply_functions
        df = self.apply_functions_on_df(df, apply_functions)

        if use_table_schema:
            df = self.convert_df_from_aito_table_schema(df, use_table_schema)

        if out_format != in_format or convert_options:
            self.df_to_format(df, out_format, write_output, convert_options)

        if create_table_schema:
            schema = self.schema_handler.generate_table_schema_from_pandas_dataframe(df)
            with create_table_schema.open(mode='w') as f:
                json.dump(schema, f, indent=4, sort_keys=True)




import datetime
import logging
from pathlib import Path


def root_path():
    return Path(__file__).parent.parent.parent


def set_up_logger(log_file_path: Path = None, logging_level: int = logging.INFO):
    """
    Wrapper for logging basic config
    :param log_file_path: path to log file. if not defined then use stderr
    :param logging_level: The logging levels
    """
    if log_file_path:
        log_file_path.parent.mkdir(parents=True, exist_ok=True)
        if log_file_path.exists():
            log_file_name = log_file_path.name + str(datetime.datetime.now().isoformat(' ', 'seconds'))
            log_file_path = str(log_file_path.parent / log_file_name)

    logging.basicConfig(filename=log_file_path, level=logging_level,
                        format='%(asctime)-5s %(name)-5s %(levelname)-10s %(message)s',
                        datefmt='%H:%M:%S')


import argparse
import sys
from pathlib import Path
from abc import abstractmethod


class AitoArgParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write(f"error: {message}\n")
        self.print_help()
        sys.exit(2)

    def check_valid_path(self, str_path, check_exists=False):
        try:
            path = Path(str_path)
        except Exception as e:
            self.error(f"invalid path {str_path}")
            raise e
        if check_exists and not path.exists():
            self.error(f"path {str_path} does not exist")
        return path

    def parse_input_arg_value(self, input_arg: str):
        return sys.stdin if input_arg == '-' else self.check_valid_path(input_arg, True)

    def parse_output_arg_value(self, output_arg: str):
        return sys.stdout if output_arg == '-' else self.check_valid_path(output_arg)

    @staticmethod
    def ask_confirmation(content, default: bool = None) -> bool:
        valid_responses = {
            'yes': True,
            'y': True,
            'no': False,
            'n': False
        }
        if not default:
            prompt = '[y/n]'
        elif default:
            prompt = '[Y/n]'
        else:
            prompt= '[y/N]'
        while True:
            sys.stdout.write(f"{content} {prompt}")
            response = input().lower()
            if default and response == '':
                return default
            elif response in valid_responses:
                return valid_responses[response]
            else:
                sys.stdout.write("Please respond with yes(y) or no(n)'\n")


class ParserWrapper():
    def __init__(self, add_help=True):
        if add_help:
            self.parser = AitoArgParser(formatter_class=argparse.RawTextHelpFormatter)
        else:
            self.parser = AitoArgParser(formatter_class=argparse.RawTextHelpFormatter, add_help=False)

    @abstractmethod
    def parse_and_execute(self, parsing_argsw):
        pass


import argparse
import json
import sys

from aito.cli.parser import AitoArgParser
from aito.utils.data_frame_handler import DataFrameHandler
from aito.utils.schema_handler import SchemaHandler


def add_infer_format_parser(format_subparsers, format_name):
    format_parser = format_subparsers.add_parser(format_name,
                                                 help=f"infer a table schema from a {format_name} file")
    format_parser.formatter_class = argparse.RawTextHelpFormatter
    # add share arguments between formats
    format_parser.add_argument('-e', '--encoding', type=str, default='utf-8',
                               help="encoding to use (default: 'utf-8')")
    format_parser.add_argument('input', default='-', type=str, nargs='?',
                               help="input file (when no input or when input is -, read standard input)")
    return format_parser


def add_infer_csv_parser(format_subparsers):
    parser = add_infer_format_parser(format_subparsers, 'csv')
    parser.description = 'infer a table schema from an excel file. '
    parser.add_argument('-d', '--delimiter', type=str, default=',',
                        help="delimiter to use. Need escape (default: ',')")
    parser.add_argument('-p', '--decimal', type=str, default='.',
                        help="Character to recognize decimal point (default '.')")
    parser.epilog = '''Example:
  aito infer-table-schema csv myFile.csv
  aito infer-table-schema csv -d ';' < mySemicolonDelimiterFile.csv > inferredSchema.json
  '''
    return parser


def add_infer_excel_parser(format_subparsers):
    parser = add_infer_format_parser(format_subparsers, 'excel')
    parser.description = '''Infer table schema from an excel file, accept both xls and xlsx.
If the file has multiple sheets, read the first sheet by default
    '''
    parser.add_argument('-o', '--one-sheet', type=str, metavar='sheet-name',
                        help='read a sheet by sheet name')
    parser.epilog = '''Example:
  aito infer-table-schema excel ./myFile.xls
  aito infer-table-schema excel -o firstSheet < myFile.xslx
  '''
    return parser


def add_infer_json_parser(format_subparsers):
    parser = add_infer_format_parser(format_subparsers, 'json')
    return parser


def add_infer_ndjson_parser(format_subparsers):
    parser = add_infer_format_parser(format_subparsers, 'ndjson')
    return parser


def add_infer_table_schema_parser(action_subparsers):
    """
    :param action_subparsers: Action subparsers from the main parser
    :return:
    """
    infer_parser = action_subparsers.add_parser('infer-table-schema', help='infer Aito table schema from a file')
    infer_parser.formatter_class = argparse.RawTextHelpFormatter
    infer_parser.epilog = '''To see help for a specific format:
  aito infer-table-schema <input-format> - h

When no input or when input is -, read standard input.
You must use input file instead of standard input for excel file

Example:
  aito infer-table-schema json myFile.json > inferredSchema.json
  aito infer-table-schema csv myFile.csv > inferredSchema.json
  aito infer-table-schema excel < myFile.xlsx > inferredSchema.json
  '''

    format_sub_parsers = infer_parser.add_subparsers(title='input-format',
                                                     description='infer from a specific format',
                                                     parser_class=AitoArgParser,
                                                     dest='input-format',
                                                     metavar="<input-format>")
    format_sub_parsers.required=True

    add_infer_csv_parser(format_sub_parsers)
    add_infer_excel_parser(format_sub_parsers)
    add_infer_json_parser(format_sub_parsers)
    add_infer_ndjson_parser(format_sub_parsers)


def execute_infer_table_schema(main_parser, parsed_args):
    in_format = parsed_args['input-format']
    read_args = {
        'read_input': main_parser.parse_input_arg_value(parsed_args['input']),
        'in_format': in_format,
        'read_options': {
            'encoding': parsed_args['encoding']
        }
    }

    if in_format == 'csv':
        read_args['read_options']['delimiter'] = parsed_args['delimiter']
        read_args['read_options']['decimal'] = parsed_args['decimal']
    elif in_format == 'excel':
        if parsed_args['input'] == '-':
            main_parser.error('Use file path instead of standard input for excel file')
        if parsed_args['one_sheet']:
            read_args['read_options']['sheet_name'] = parsed_args['one_sheet']

    df = DataFrameHandler().read_file_to_df(**read_args)
    inferred_schema = SchemaHandler().generate_table_schema_from_pandas_dataframe(df)
    json.dump(inferred_schema, sys.stdout, indent=4, sort_keys=True)
    return 0



import argparse
import logging
import sys

import argcomplete

from aito.cli.convert_parser import add_convert_parser, execute_convert
from aito.cli.infer_table_schema_parser import add_infer_table_schema_parser, execute_infer_table_schema
from aito.cli.database_parser import add_database_parser, execute_database_operation
from aito.cli.parser import AitoArgParser


class MainParser:
    def __init__(self):
        self.parser = AitoArgParser(formatter_class=argparse.RawTextHelpFormatter, prog="aito")
        action_subparsers = self.parser.add_subparsers(title='action',
                                                       description='action to perform',
                                                       dest='action',
                                                       parser_class=AitoArgParser,
                                                       metavar="<action>")
        action_subparsers.required=True
        add_infer_table_schema_parser(action_subparsers)
        add_convert_parser(action_subparsers)
        add_database_parser(action_subparsers)
        argcomplete.autocomplete(self.parser)

    def parse_and_execute(self, parsing_args):
        parsed_args = vars(self.parser.parse_args(parsing_args))
        action = parsed_args['action']
        if action == 'infer-table-schema':
            execute_infer_table_schema(self.parser, parsed_args)
        elif action == 'convert':
            execute_convert(self.parser, parsed_args)
        elif action == 'database':
            execute_database_operation(self.parser, parsed_args)
        return 0


def main():
    logging.basicConfig(level=logging.INFO, format='%(asctime)-5s %(name)-5s %(levelname)-10s %(message)s',
                        datefmt='%H:%M:%S')
    main_parser = MainParser()
    main_parser.parse_and_execute(sys.argv[1:])


if __name__ == '__main__':
    main()



import asyncio
import logging
from pathlib import Path
from typing import Dict, List
import time
from timeit import default_timer

import requests
from aiohttp import ClientSession


class ClientError(Exception):
    def __init__(self, message):
        super().__init__(message)


class AitoClient:
    request_methods = {
        'PUT': requests.put, 'POST': requests.post, 'GET': requests.get, 'DELETE': requests.delete
    }

    query_api_paths = [
        '/api/v1/_search',
        '/api/v1/_predict',
        '/api/v1/_recommend',
        '/api/v1/_evaluate',
        '/api/v1/_similarity',
        '/api/v1/_match',
        '/api/v1/_relate',
        '/api/v1/_query'
    ]

    database_api_prefix_and_methods = {
        '/api/v1/schema': ['GET', 'PUT', 'DELETE'],
        '/api/v1/data': ['POST', 'GET'],
        '/api/v1/data/_delete': ['POST']
    }

    rw_key_paths = list(database_api_prefix_and_methods.keys())
    ro_key_paths = query_api_paths + ['/version']

    def __init__(self, url: str, rw_key: str, ro_key: str):
        self.logger = logging.getLogger("AitoClient")
        if not url:
            raise ClientError("Aito instance url is required")
        if not (url.startswith('https://') and url.endswith('.api.aito.ai')):
            raise ClientError(f"Invalid instance url: {url}. The instance url should look similar "
                              f"to https://instance-name.api.aito.ai")
        self.url = url

        if not rw_key and not ro_key:
            raise ClientError("Both read-write and read-only key are missing")

        self.rw_key = rw_key
        self.ro_key = ro_key
        if not rw_key:
            self.logger.warning("Read-write key is not defined. APIs which require read-write key will fail")
        if not ro_key:
            self.logger.warning(f"Read-only key is not defined. Use read-write key instead")
            self.ro_key = self.rw_key

    def check_path_and_build_headers(self, path: str):
        """
        Check if path is validate aito path.
        Use read-write or read-only key accordingly
        :param path:
        :return:
        """
        headers = {'Content-Type': 'application/json'}
        is_database_api_path = any([path.startswith(db_path) for db_path in self.database_api_prefix_and_methods])
        if is_database_api_path:
            if not self.rw_key:
                raise ClientError(f"Path {path} requires read-write key")
            headers['x-api-key'] = self.rw_key
        elif path in self.ro_key_paths:
            headers['x-api-key'] = self.ro_key
        elif path.startswith('/api/v1/'):
            self.logger.warning(f"Unrecognized path {path}")
            if self.rw_key:
                headers['x-api-key'] = self.rw_key
            else:
                headers['x-api-key'] = self.ro_key
        else:
            raise ClientError(f"invalid path {path}")
        return headers

    async def fetch(self, session: ClientSession, req_method: 'str', path: str, json_data: Dict):
        self.logger.debug(f"{req_method} to {path} with {str(json_data)[:250]}...")
        headers = self.check_path_and_build_headers(path)
        async with session.request(method=req_method, url=self.url + path, json=json_data, headers=headers) as resp:
            json_resp = await resp.json()
            self.logger.debug(f"got response ${str(json_resp)[:250]}...")
            return json_resp

    def async_same_requests(self, request_count: int, request_method: str, path: str, json_data: Dict = None):
        """
        Run async n similar requests
        :param request_count: number of requests
        :param request_method:
        :param path: aito API endpoints path
        :param json_data: request content
        :return:
        """
        async def run():
            async with ClientSession() as session:
                tasks = [self.fetch(session, request_method, path, json_data)] * request_count
                return await asyncio.gather(*tasks)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop.run_until_complete(run())

    def async_requests(self, request_methods: List[str], request_paths: List[str], request_data: List[Dict]):
        """
        Run async multiple requests
        :param request_methods: list of request methods
        :param request_paths: list of request API path
        :param request_data: list of request data
        :return:
        """
        async def run():
            async with ClientSession() as session:
                tasks = [self.fetch(session, method, request_paths[idx], request_data[idx])
                         for idx, method in enumerate(request_methods)]
                return await asyncio.gather(*tasks)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop.run_until_complete(run())

    def request(self, req_method: str, path: str, data=None):
        headers = self.check_path_and_build_headers(path)
        url = self.url + path

        logger = self.logger
        logger.debug(f"{req_method} to {path} with data {str(data)[:250]}...")
        try:
            r = self.request_methods[req_method](url, headers=headers, json=data)
            r.raise_for_status()
        except requests.HTTPError as e:
            raise ClientError(f"{e}\n{r.text}")
        except requests.exceptions.RequestException as e:
            raise ClientError(str(e))
        except Exception as e:
            self.logger.exception(f"Unknown error {e}")
            raise e
        return r.json()

    def get_version(self):
        return self.request('GET', '/version')

    def put_database_schema(self, database_schema):
        self.logger.info("Creating database schema...")
        r = self.request('PUT', '/api/v1/schema', database_schema)
        self.logger.info("Database schema created")
        return r

    def get_database_schema(self):
        return self.request('GET', '/api/v1/schema')

    def delete_database(self):
        self.logger.info("Deleting the whole database...")
        r = self.request('DELETE', '/api/v1/schema')
        self.logger.info("Database deleted")
        return r

    def put_table_schema(self, table_name, table_schema):
        self.logger.info(f"Creating table '{table_name}' schema...")
        r = self.request('PUT', f"/api/v1/schema/{table_name}", table_schema)
        self.logger.info(f"Table '{table_name}' schema created")
        return r

    def delete_table(self, table_name):
        self.logger.info(f"Deleting table '{table_name}'")
        r = self.request('DELETE', f"/api/v1/schema/{table_name}")
        self.logger.info(f"Table '{table_name}' deleted")
        return r

    def get_table_schema(self, table_name):
        return self.request('GET', f"/api/v1/schema/{table_name}")

    def get_existing_tables(self) -> List[str]:
        return list(self.get_database_schema()['schema'].keys())

    def check_table_existed(self, table_name) -> bool:
        return table_name in self.get_existing_tables()

    def populate_table_entries(self, table_name, entries):
        if len(entries) > 1000:
            self.populate_table_entries_by_batches(table_name, entries)
        else:
            self.logger.info(f"Start uploading {len(entries)} entries to table '{table_name}'...")
            self.request('POST', f"/api/v1/data/{table_name}/batch", entries)
            self.logger.info(f"Uploaded {len(entries)} entries to table '{table_name}'")

    def populate_table_entries_by_batches(self, table_name, entries, batch_size=1000):
        self.logger.info(f"Start uploading {len(entries)} entries to table '{table_name}' "
                         f"with batch size of {batch_size}...")
        begin_idx = 0
        populated = 0
        while begin_idx < len(entries):
            last = begin_idx + batch_size
            entries_batch = entries[begin_idx:last]
            try:
                self.populate_table_entries(table_name, entries_batch)
                populated += len(entries_batch)
            except Exception as e:
                self.logger.error(f"Batch {begin_idx}:{last} failed: {e}")
            begin_idx = last

        self.logger.info(f"Uploaded {populated}/{len(entries)} entries to table '{table_name}'")

    def populate_table_by_file_upload(self, table_name: str, file_path: Path):
        if file_path.suffixes[-2:] != ['.ndjson', '.gz']:
            raise ClientError("Cannot upload file. Uploading file must be in gzip compressed ndjson format")
        self.logger.info("Initiating file upload...")
        r = self.request('POST', f"/api/v1/data/{table_name}/file")
        upload_session_id = r['id']
        session_end_point = f'/api/v1/data/{table_name}/file/{upload_session_id}'
        s3_url = r['url']
        upload_req_method = r['method']

        self.logger.info("Uploading file to S3...")
        try:
            requests.request(upload_req_method, s3_url, data=file_path.open(mode='rb'))
            self.logger.info("Uploading file to S3 completed")
        except Exception as e:
            self.logger.error(f"Failed to upload file to S3: {e}")
            raise e
        self.logger.info("Triggering file processing...")
        self.request('POST', session_end_point)

        start_polling = default_timer()
        while True:
            processing_progress = self.request('GET', session_end_point)
            status = processing_progress['status']
            self.logger.info(f"completed count: {status['completedCount']}, throughput: {status['throughput']}")
            if processing_progress['errors']['message'] != 'Last 0 failing rows':
                self.logger.error(processing_progress['errors'])
            if status['finished']:
                break
            time_elapsed = default_timer() - start_polling
            if time_elapsed < 60:
                sleeping_time = 5
            elif time_elapsed < 180:
                sleeping_time = 15
            elif time_elapsed < 300:
                sleeping_time = 30
            else:
                sleeping_time = 60
            time.sleep(sleeping_time)

        self.logger.info(f"Populate table '{table_name}' by file upload completed")

    def query_table_entries(self, table_name: str, limit: int = None):
        query = {'from': table_name}
        if limit:
            query['limit'] = limit
        return self.request('POST', '/api/v1/_query', query)







import logging
from typing import List, Dict

import pandas as pd
from langdetect import detect


class SchemaHandler:
    def __init__(self):
        self.logger = logging.getLogger("SchemaHandler")
        self.pandas_dtypes_name_to_aito_type = {
            'string': 'Text',
            'unicode': 'Text',
            'bytes': 'Text',
            'floating': 'Decimal',
            'integer': 'Int',
            'mixed - integer': 'Decimal',
            'mixed - integer - float': 'Decimal',
            'decimal': 'Decimal',
            'complex': 'Decimal',
            'categorical': 'String',
            'boolean': 'Boolean',
            'datetime64': 'String',
            'datetime': 'String',
            'date': 'String',
            'timedelta64': 'String',
            'timedelta': 'String',
            'time': 'String',
            'period': 'String',
            'mixed': 'Text'
        }
        self.aito_types_to_pandas_dtypes = {
            'Boolean': ['bool'],
            'Decimal': ['float64', 'float16', 'float32', 'float128'],
            'Int': ['int32', 'int8', 'int16', 'int64'],
            'String': ['str'],
            'Text': ['str']
        }

        self.lang_detect_code_to_aito_code = {
            'ko': 'cjk',
            'ja': 'cjk',
            'zh-cn': 'cjk',
            'zh-tw': 'cjk',
            'pt': 'pt-br'
        }

        self.supported_alias_analyzer = ['standard', 'whitespace', 'ar', 'hy', 'eu', 'pt-br', 'bg', 'ca', 'cjk', 'cs',
                                         'da', 'nl', 'en', 'fi', 'fr', 'gl', 'de', 'el', 'hi', 'hu', 'id', 'ga', 'it',
                                         'lv', 'no', 'fa', 'pt', 'ro', 'ru', 'es', 'sv', 'th', 'tr']
        self.sample_size = 100000

    def infer_aito_types_from_values(self, values: List):
        dtypes = pd.api.types.infer_dtype(values, skipna=True)
        return self.pandas_dtypes_name_to_aito_type[dtypes]

    def generate_table_schema_from_pandas_dataframe(self, table_df: pd.DataFrame):
        """
        Return aito schema in dictionary format
        :param table_df: The pandas DataFrame containing table data
        :return: Aito Table Schema as dict
        """
        self.logger.info("Start inferring schema...")

        rows_count = table_df.shape[0]

        columns_schema = {}
        for col in table_df.columns.values:
            if rows_count >= self.sample_size:
                col_df = table_df[col].sample(self.sample_size)
            else:
                col_df = table_df[col]
            col_data = col_df.values
            col_aito_type = self.infer_aito_types_from_values(col_data)
            col_na_count = col_df.isna().sum()
            col_schema = {
                'nullable': True if col_na_count > 0 else False,
                'type': col_aito_type
            }
            if col_schema['type'] == 'Text':
                col_unique_val_count = col_df.nunique()
                col_non_na_count = rows_count - col_na_count
                if (col_non_na_count / col_unique_val_count) > 2:
                    col_schema['type'] = 'String'
                else:
                    col_text = col_df.str.cat(sep=' ')
                    try:
                        detected_lang = detect(col_text)
                        if detected_lang in self.lang_detect_code_to_aito_code:
                            detected_lang = self.lang_detect_code_to_aito_code[detected_lang]
                    except:
                        detected_lang = None
                    if detected_lang and detected_lang in self.supported_alias_analyzer:
                        col_schema['analyzer'] = detected_lang

            columns_schema[col] = col_schema

        table_schema = {'type': 'table', 'columns': columns_schema}
        self.logger.info("Finished inferring schema")
        return table_schema

    def validate_table_schema(self, table_schema: Dict):
        """

        :param table_schema: table schema as dict
        :return:
        """
        def validate_required_arg(object_name: str, object_dict: Dict, arg_name: str):
            if arg_name not in object_dict:
                raise ValueError(f"{object_name} missing '{arg_name}'")

        def validate_arg_type(object_name: str, object_dict: Dict, arg_name: str, arg_type: object,
                              accepted_values: List = None):
            arg_value = object_dict[arg_name]
            if not isinstance(arg_value, arg_type):
                raise ValueError(f"Invalid {object_name} {arg_name} type")
            if accepted_values:
                if arg_value not in accepted_values:
                    accepted_values_str = accepted_values[0] if len(accepted_values) == 1 \
                        else f"one of {accepted_values}"
                    raise ValueError(f"{object_name} '{arg_name}' must be {accepted_values_str}")

        def validate_arg(object_name: str, object_dict: Dict, arg_name: str, required: bool = False,
                         arg_type: object = None, accepted_values: List = None):
            if required:
                validate_required_arg(object_name, object_dict, arg_name)
            if arg_name in object_dict and arg_type:
                validate_arg_type(object_name, object_dict, arg_name, arg_type, accepted_values)

        self.logger.info("Start validating schema...")
        validate_arg('table', table_schema, 'type', True, str, ['table'])
        validate_arg('table', table_schema, 'columns', True, dict)
        for col in table_schema['columns']:
            col_schema = table_schema['columns'][col]
            validate_arg(col, col_schema, 'type', True, str, list(self.aito_types_to_pandas_dtypes.keys()))
            validate_arg(col, col_schema, 'nullable', False, bool)
        self.logger.info("Finished validating schema")
        return table_schema



import argparse
import json
import os
import sys
from datetime import datetime

from dotenv import load_dotenv

from aito.utils.aito_client import AitoClient
from aito.cli.parser import AitoArgParser
from aito.utils.data_frame_handler import DataFrameHandler


def create_client_from_parsed_args(main_parser, parsed_args):
    def get_env_variable(variable_name):
        if variable_name not in env_var:
            main_parser.error(f"{variable_name} env variable not found")
        return env_var[variable_name]

    if parsed_args['use_env_file']:
        env_file_path = main_parser.check_valid_path(parsed_args['use_env_file'], True)
        load_dotenv(env_file_path)
    env_var = os.environ

    client_args = {
        'url': get_env_variable('AITO_INSTANCE_URL') if parsed_args['url'] == '.env' else parsed_args['url'],
        'rw_key': get_env_variable('AITO_RW_KEY') if parsed_args['read_write_key'] == '.env'
        else parsed_args['read_write_key'],
        'ro_key': get_env_variable('AITO_RO_KEY') if parsed_args['read_only_key'] == '.env'
        else parsed_args['read_only_key']
    }
    return AitoClient(**client_args)


def add_quick_add_table_parser(operation_subparsers):
    parser = operation_subparsers.add_parser(
        'quick-add-table', help="infer schema, create table from file, and upload to the database"
    )
    parser.epilog = '''Note: this operation requires write permission to create a ndjson file for file upload
Example:
  aito database quick-add-table myFile.json
  aito database quick-add-table --table-name aTableName myFile.csv
  aito database quick-add-table -n myTable -f csv myFile
'''
    parser.add_argument(
        '-n', '--table-name', type=str,
        help='create a table with the given name (default: the input file name without the extension)')

    file_format_choices = ['infer'] + DataFrameHandler.allowed_format
    parser.add_argument(
        '-f', '--file-format', type=str, choices=file_format_choices, default='infer',
        help='specify the input file format (default: infer from the file extension)')
    parser.add_argument('input-file', type=str, help="path to the input file")
    return parser


def execute_quick_add_table(main_parser, parsed_args):
    input_file_path = main_parser.check_valid_path(parsed_args['input-file'])
    table_name = parsed_args['table_name'] if parsed_args['table_name'] else input_file_path.stem
    in_format = input_file_path.suffixes[0].replace('.', '') if parsed_args['file_format'] == 'infer' \
        else parsed_args['file_format']

    now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    converted_file_path = input_file_path.parent / f"{input_file_path.stem}_{now}.ndjson.gz"
    schema_path = input_file_path.parent / f"{table_name}_schema_{now}.json"
    convert_options = {
        'read_input': input_file_path,
        'write_output': converted_file_path,
        'in_format': in_format,
        'out_format': 'ndjson',
        'convert_options': {'compression': 'gzip'},
        'create_table_schema': schema_path
    }
    df_handler = DataFrameHandler()
    df_handler.convert_file(**convert_options)

    client = create_client_from_parsed_args(main_parser, parsed_args)
    with schema_path.open() as f:
        inferred_schema = json.load(f)
    client.put_table_schema(table_name, inferred_schema)
    client.populate_table_by_file_upload(table_name, converted_file_path)

    if converted_file_path.exists():
        converted_file_path.unlink()
    schema_path.unlink()
    return 0


def add_create_table_parser(operation_subparsers):
    parser = operation_subparsers.add_parser('create-table', help='create a table using the input table schema')
    parser.epilog = '''With no table-schema-input or when input is -, read from standard input
Example:
  aito client create-table aTableName < path/to/tableSchemaFile.json
  '''
    parser.add_argument('table-name', type=str, help="name of the table to be created")
    parser.add_argument('schema-input', default='-', type=str, nargs='?', help="table schema input")
    return parser


def execute_create_table(main_parser, parsed_args):
    client = create_client_from_parsed_args(main_parser, parsed_args)
    table_name = parsed_args['table-name']
    if parsed_args['schema-input'] == '-':
        table_schema = json.load(sys.stdin)
    else:
        input_path = main_parser.check_valid_path(parsed_args['schema-input'])
        with input_path.open() as f:
            table_schema = json.load(f)
    client.put_table_schema(table_name, table_schema)
    return 0


def add_delete_table_parser(operation_subparsers):
    parser = operation_subparsers.add_parser('delete-table',
                                             help="delete a table schema and all content inside the table")
    parser.add_argument('table-name', type=str, help="name of the table to be deleted")
    return parser


def execute_delete_table(main_parser, parsed_args):
    client = create_client_from_parsed_args(main_parser, parsed_args)
    table_name = parsed_args['table-name']
    if main_parser.ask_confirmation(f"Confirm delete table '{table_name}'? The action is irreversible", False):
        client.delete_table(table_name)
    return 0


def add_delete_database_parser(operation_subparsers):
    parser = operation_subparsers.add_parser('delete-database', help='delete the whole database')
    return parser


def execute_delete_database(main_parser, parsed_args):
    client = create_client_from_parsed_args(main_parser, parsed_args)
    if main_parser.ask_confirmation(f"Confirm delete the whole database? The action is irreversible", False):
        client.delete_database()
    return 0


def add_upload_batch_parser(operation_subparsers):
    parser = operation_subparsers.add_parser('upload-batch', help='populate table entries to a table')
    parser.epilog = '''With no input, or when input is -, read table content from standard input
Example:
  aito client upload-batch myTable path/to/myTableEntries.json
  aito client upload-batch myTable < path/to/myTableEntries.json
  '''
    parser.add_argument('table-name', type=str, help='name of the table to be populated')
    parser.add_argument('input', default='-', type=str, nargs='?', help='input file or stream in JSON array format')


def execute_upload_batch(main_parser, parsed_args):
    client = create_client_from_parsed_args(main_parser, parsed_args)
    table_name = parsed_args['table-name']
    if parsed_args['input'] == '-':
        table_content = json.load(sys.stdin)
    else:
        input_path = main_parser.check_valid_path(parsed_args['input'])
        with input_path.open() as f:
            table_content = json.load(f)
    client.populate_table_entries(table_name, table_content)
    return 0


def add_upload_file_parser(operation_subparsers):
    parser = operation_subparsers.add_parser('upload-file', help='populate a file to a table')
    parser.epilog = '''Example:
  aito client upload-file tableName path/to/myFile.csv
  aito client upload-file -f csv tableName path/to/myFile.txt
  '''

    parser.add_argument('table-name', type=str, help='name of the table to be populated')
    parser.add_argument('input-file', type=str, help="path to the input file")
    file_format_choices = ['infer'] + DataFrameHandler.allowed_format
    parser.add_argument('-f', '--file-format', type=str, choices=file_format_choices,
                        default='infer', help='specify input file format (default: infer from the file extension)')


def execute_upload_file(main_parser, parsed_args):
    client = create_client_from_parsed_args(main_parser, parsed_args)
    table_name = parsed_args['table-name']

    if not client.check_table_existed(table_name):
        main_parser.error(f"Table '{table_name}' does not exist. Please create table first.")

    input_file_path = main_parser.check_valid_path(parsed_args['input-file'])

    in_format = input_file_path.suffixes[0].replace('.', '') if parsed_args['file_format'] == 'infer' \
        else parsed_args['file_format']

    now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    converted_file_path = input_file_path.parent / f"{input_file_path.stem}_{now}.ndjson.gz"
    convert_options = {
        'read_input': input_file_path,
        'write_output': converted_file_path,
        'in_format': in_format,
        'out_format': 'ndjson',
        'convert_options': {'compression': 'gzip'},
        'use_table_schema': client.get_table_schema(table_name)
    }
    df_handler = DataFrameHandler()
    df_handler.convert_file(**convert_options)
    client.populate_table_by_file_upload(table_name, converted_file_path)

    if converted_file_path.exists():
        converted_file_path.unlink()
    return 0


def add_database_parser(action_subparsers):
    database_parser = action_subparsers.add_parser('database',
                                                   help='perform operations with your Aito database instance')
    database_parser.formatter_class = argparse.RawTextHelpFormatter
    database_parser.epilog = '''You must provide credentials to execute database operations
If no credential options is given, use the credentials from environment variable by defaults.
The AITO_INSTANCE_URL should look similar to https://my-instance.api.aito.ai

To see help for a specific operation:
  aito database <operation> -h  

Example:
  aito database quick-add-table myTable.csv
  aito database create-table tableName < path/to/tableSchema
  aito database -e path/to/myCredentials.env upload-file myTable path/to/myFile
  aito database -u MY_AITO_INSTANCE_URL -r MY_RO_KEY -w MY_RW_KEY upload-batch myTable < path/to/myTableEntries
    '''
    credential_args = database_parser.add_argument_group("optional credential arguments")
    credential_args.add_argument('-e', '--use-env-file', type=str, metavar='env-input-file',
                                 help='set up the client using a .env file containing the required env variables')
    credential_args.add_argument('-r', '--read-only-key', type=str, default='.env',
                                 help='specify aito read-only API key')
    credential_args.add_argument('-u', '--url', type=str, default='.env', help='specify aito instance url')
    credential_args.add_argument('-w', '--read-write-key', type=str, default='.env',
                                 help='specify aito read-write API key')

    operation_subparsers = database_parser.add_subparsers(title="operation",
                                                          description="operation to perform",
                                                          parser_class=AitoArgParser,
                                                          dest="operation",
                                                          metavar="<operation>")
    operation_subparsers.required=True

    add_quick_add_table_parser(operation_subparsers)
    add_create_table_parser(operation_subparsers)
    add_delete_table_parser(operation_subparsers)
    add_delete_database_parser(operation_subparsers)
    add_upload_batch_parser(operation_subparsers)
    add_upload_file_parser(operation_subparsers)


def execute_database_operation(main_parser, parsed_args):
    if parsed_args['operation'] == 'quick-add-table':
        execute_quick_add_table(main_parser, parsed_args)
    elif parsed_args['operation'] == 'create-table':
        execute_create_table(main_parser, parsed_args)
    elif parsed_args['operation'] == 'delete-table':
        execute_delete_table(main_parser, parsed_args)
    elif parsed_args['operation'] == 'delete-database':
        execute_delete_database(main_parser, parsed_args)
    elif parsed_args['operation'] == 'upload-batch':
        execute_upload_batch(main_parser, parsed_args)
    elif parsed_args['operation'] == 'upload-file':
        execute_upload_file(main_parser, parsed_args)

    return 0



from setuptools import setup, find_packages
import os

install_requires = [
    'python-dotenv',
    'requests',
    'aiohttp',
    'pandas',
    'ndjson',
    'xlrd',
    'langdetect',
    'argcomplete'
]

if os.environ.get('CONVERT_README'):
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
else:
    long_description = ''

VERSION = "0.0.4"

setup(
    name='aitoai',
    version=VERSION,
    author='aito.ai',
    author_email='admin@aito.ai',
    description='A collection of python support tools and scripts for Aito.ai',
    long_description=long_description,
    url='https://github.com/AitoDotAI/aito-python-tools',
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points={'console_scripts': ['aito = aito.cli.main_parser:main']}
)



import argparse
import json
import sys

from aito.utils.data_frame_handler import DataFrameHandler
from aito.cli.parser import AitoArgParser


def add_convert_format_parser(format_subparsers, format_name):
    format_parser = format_subparsers.add_parser(format_name,
                                                 help=f"convert a {format_name} into ndjson|json format")
    format_parser.formatter_class = argparse.RawTextHelpFormatter
    # add share arguments between formats
    either_use_or_create_schema = format_parser.add_mutually_exclusive_group()
    either_use_or_create_schema.add_argument('-c', '--create-table-schema', metavar='schema-output-file',
                                             type=str,
                                             help='create an inferred aito schema and write to output file')
    either_use_or_create_schema.add_argument('-s', '--use-table-schema', metavar='schema-input-file', type=str,
                                             help='convert the data to match the input table schema')
    format_parser.add_argument('-e', '--encoding', type=str, default='utf-8',
                               help="encoding to use (default: 'utf-8')")
    format_parser.add_argument('-j', '--json', action='store_true', help='convert to json format')

    format_parser.add_argument('input', default='-', type=str, nargs='?',
                               help="input file (when no input or when input is -, read standard input)")
    return format_parser


def add_convert_csv_parser(format_subparsers):
    parser = add_convert_format_parser(format_subparsers, 'csv')
    parser.add_argument('-d', '--delimiter', type=str, default=',', help="delimiter to use (default: ',')")
    parser.add_argument('-p', '--decimal', type=str, default='.',
                        help="character to recognize decimal point (default '.')")
    parser.epilog = '''Example:
  aito convert csv myFile.csv
  aito convert csv -d ';' --json < mySemicolonDelimiterFile.csv > convertedFile.json
  '''
    return parser


def add_convert_excel_parser(format_subparsers):
    parser = add_convert_format_parser(format_subparsers, 'excel')
    parser.description = 'Convert excel format input, accept both xls and xlsx. ' \
                                       'Read the first sheet of the file by default'
    parser.add_argument('-o', '--one-sheet', type=str, metavar='sheet-name',
                                      help='read a sheet of the excel file')
    parser.epilog = '''Example:
    aito convert excel ./myFile.xls
    aito convert excel -o firstSheet myMultipleSheetsFile.xslx
    '''
    return parser


def add_convert_json_parser(format_subparsers):
    parser = add_convert_format_parser(format_subparsers, 'json')
    return parser


def add_convert_ndjson_parser(format_subparsers):
    parser = add_convert_format_parser(format_subparsers, 'ndjson')
    return parser


def add_convert_parser(action_subparsers):
    """
    :param action_subparsers: Action subparsers from the main parser
    :return:
    """
    convert_parser = action_subparsers.add_parser('convert', help='convert a file into ndjson|json format')
    convert_parser.formatter_class = argparse.RawTextHelpFormatter
    convert_parser.epilog = '''To see help for a specific format:
  aito convert <input-format> - h

When no input or when input is -, read standard input. 
You must use input file instead of standard input for excel file

Example:
  aito convert json myFile.json > convertedFile.ndjson
  aito convert csv -c myInferredTableSchema.json --json myFile.csv > convertedFile.json
  aito convert excel -s desiredSchema.json < myFile.xlsx > convertedFile.ndjson
  '''

    format_subparsers = convert_parser.add_subparsers(title='input-format',
                                                      description='convert from a specific format',
                                                      parser_class=AitoArgParser,
                                                      dest='input-format',
                                                      metavar='<input-format>')
    format_subparsers.required=True
    add_convert_csv_parser(format_subparsers)
    add_convert_excel_parser(format_subparsers)
    add_convert_json_parser(format_subparsers)
    add_convert_ndjson_parser(format_subparsers)


def execute_convert(main_parser, parsed_args):
    in_format = parsed_args['input-format']
    convert_args = {
        'read_input': main_parser.parse_input_arg_value(parsed_args['input']),
        'write_output': sys.stdout,
        'in_format': parsed_args['input-format'],
        'out_format': 'json' if parsed_args['json'] else 'ndjson',
        'read_options': {
            'encoding': parsed_args['encoding']
        },
        'convert_options': {},
        'create_table_schema':
            main_parser.check_valid_path(parsed_args['create_table_schema'])
            if parsed_args['create_table_schema'] else None,
    }

    if parsed_args['use_table_schema']:
        schema_path = main_parser.check_valid_path(parsed_args['use_table_schema'], check_exists=True)
        with schema_path.open() as f:
            table_schema = json.load(f)
        convert_args['use_table_schema'] = table_schema

    if in_format == 'csv':
        convert_args['read_options']['delimiter'] = parsed_args['delimiter']
        convert_args['read_options']['decimal'] = parsed_args['decimal']

    if in_format == 'excel':
        if parsed_args['input'] == '-':
            main_parser.error('Use file path instead of standard input for excel file')
        if parsed_args['one_sheet']:
            convert_args['read_options']['sheet_name'] = parsed_args['one_sheet']

    DataFrameHandler().convert_file(**convert_args)
    return 0
